import math

import matplotlib.pylab as plt
import matplotlib.pyplot as plt
import pandas as pd

from face2face.imports.load_all_data import Data
from face2face.statistics.null_modell import configuration_model_label_z_score_mixing_matrix
from face2face.statistics.null_modell import shuffle_label_z_score_mixing_matrix


def plot_diff_scales(ax1, x_value, y_value, x2_value, y2_value, title):

    """Creates the subplots

       Creates the subplots for every attribute combination and defines the parameters for their visualization

       Parameters
       ----------
       ax1: axes
           Basic axes parameter for the figure without any layout information yet
       x_value: list
           List with the amount of edges between the two attribute-values for every iteration of the null model
           functions.
       y_value:
           Frequency of the occurrence of the amount of edges for every iteration
       x2_value: float
           Average z-score for this attribute-value combination
       y2_value: float
           Maximum value of y_value list for a better representation of the z-score
       title: list
           Contains the two attribute-values that gets compared in this iteration

       Returns
       -------
       ax1: axes
           Parameter for the visualization of the edge-frequency between the two attribute-values
       ax2: axes
           Parameter for the visualization of the z-score for the two attribute-values
       """

    ax1.title.set_text(str(title))
    ax2 = ax1.twiny()
    ax1.bar(x2_value, y2_value, color="#0B6C11", width=0.08, align="center")
    ax1.set_ylabel('Frequency', color="k")
    ax1.set_xlim(-5, 5)
    ax1.tick_params(axis='x', labelcolor="k")
    ax2.bar(x_value, y_value, color="#23D42F", align="center")
    ax2.set_ylabel('Frequency')
    ax1.set_xlabel('z-score')

    return ax1, ax2


def plot_null_model_subplots(Data, runs, label="type", shuffle_label=False, force_simple_graph=False,
                             model="configuration_model"):

    """Creates a figure for all attribute-value combinations

       Creates and plots a figure with subplots for every attribute-value combination. Every subplot shows the frequency
       of the amount of edges for the given attribute-value combination and the average z-score.

       Parameters
       ----------
       Data: Data
           Data Object that contains Tij- and Metadata for a data set.
       runs: int
           The amount of times the function should be executed. It's a heuristic approach, so the more the runs the
           better might be the result
       label: str
           A string that tells the function for which attribute the contact matrix should be made.
       shuffle_label: bool
           Gives the option to extend the null model by randomizing the node attributes.
       force_simple_graph: bool
           Deletes parallel- and selfedges that can occur by using the networkx "configuration_model"-function if True.
       model: str
           Defines which null model should be used. Possible Values: "configuration_model", "random_identities"

       Returns
       -------
       None

       See Also
       ---------
       face2face.statistics.null_modell.configuration_model_label_z_score_mixing_matrix
       face2face.statistics.null_modell.shuffle_label_z_score_mixing_matrix
       face2face.visualization.plot_contact_matrix_heatmap.plot_cm_heatmap
       """

    if model == "configuration_model":
        contact_matrix, matrices, data_mixing_matrix, matrices2 = configuration_model_label_z_score_mixing_matrix(
            Data, runs=runs, label=label, shuffle_label=shuffle_label, force_simple_graph=force_simple_graph)

    elif model == "random_identities":
        contact_matrix, matrices2 = shuffle_label_z_score_mixing_matrix(Data, runs=runs, label=label)

    contact_matrix_list = contact_matrix.ravel().tolist()

    abs_edges_lists = []
    for i in matrices2:
        abs_edges_lists.append(i.ravel().tolist())

    abs_edges_df = pd.DataFrame(abs_edges_lists)

    frequency_df_list = []
    for column in abs_edges_df:
        frequency_df = abs_edges_df.groupby(column).size().reset_index(name="Frequency")
        frequency_df_list.append(frequency_df)

    dimension = int(math.sqrt(len(contact_matrix_list)))
    list_plot_title = []
    for i in range(dimension):
        for j in range(dimension):
            list_plot_title.append([i, j])

    i = 0
    fig, axs = plt.subplots(dimension, dimension, figsize=(5, 4))

    for row in axs:
        for ax in row:
            ax1, ax2 = plot_diff_scales(ax, frequency_df_list[i][i], frequency_df_list[i]["Frequency"],
                                        contact_matrix_list[i], max(frequency_df_list[i]["Frequency"]),
                                        list_plot_title[i])
            i += 1

    plt.tight_layout()
    plt.show()




import seaborn as sns
import matplotlib.pylab as plt

from face2face.statistics.null_modell import check_bonferroni_correction


def plot_cm_heatmap(contact_matrix):

    """Plots a heatmap for a given contact matrix

       Plots the contact matrix for a chosen attribute. The heatmap includes a bonferroni correction which is shown by
       a white * if a value is an outlier.

       Parameters
       ----------
       contact_matrix : list
           A matrixlike List of lists that contains the z-scores for a given attribute.

       Returns
       -------
       None

       References
       ------------
       .. [1] Génois, Mathieu & Zens, Maria & Lechner, Clemens & Rammstedt, Beatrice & Strohmaier, Markus. (2019).
              Building connections: How scientists meet each other during a conference.

       See Also
       ----------
       face2face.statistics.null_modell.configuration_model_label_z_score_mixing_matrix
       face2face.statistics.null_modell.shuffle_label_z_score_mixing_matrix
       face2face.visualization.plot_histogram_null_model.plot_null_model_subplots

    """
    annotation_list = check_bonferroni_correction(contact_matrix)
    ax = sns.heatmap(contact_matrix, annot=annotation_list, fmt='', cmap="bwr", linewidth=0.5)
    plt.show()

    return None


import networkx as nx
import pandas as pd

from face2face.imports.load_all_data import Data


pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)
pd.set_option('display.float_format', lambda x: '%.5f' % x)


def create_network_from_data(Data, replace_attr=False, label=None):
    r"""Allow to create a networkx Graph

        Creating a networkx graph based on the interaction data. Metadata can also be provided to add node attributes.

        Parameters
        ----------
        Data : Data
            Data Object that contains Tij- and Metadata for a dataset.
        replace_attr: bool
            If True, attributes of type string will be replaced by float numbers.
        label: optional, string
            If string label is given, the rows where label column value is NaN will be removed before creating the
            network.

        Returns
        -------
        G : NetworkX Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported dataset.

        Examples
        ---------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>                meta_attr_list=attr_list)
        >>> test_network = create_network_from_data(test_df)
        >>> print(test_network.edges)
        [(0, 1), (1, 2), (1, 3), (2, 3), (4, 6), (4, 7), (6, 7), (10, 11), (10, 12), (11, 12)]
        >>> print(test_network.nodes)
        [0, 1, 2, 3, 4, 6, 7, 10, 11, 12]
        >>> for i in test_network.nodes:
            >>> print(test_network.nodes[i])
        {'Age': 1.0, 'Sex': 1.0}
        {'Age': 0.0, 'Sex': 'NaN'}
        {'Age': 'NaN', 'Sex': 0.0}
        {'Age': 0.0, 'Sex': 1.0}
        {'Age': 'NaN', 'Sex': 0.0}
        {'Age': 1.0, 'Sex': 1.0}
        {'Age': 1.0, 'Sex': 'NaN'}
        {'Age': 2.0, 'Sex': 1.0}
        {'Age': 2.0, 'Sex': 0.0}
        {'Age': 2.0, 'Sex': 1.0}

    """
    if hasattr(Data,"metadata") is False:
        G = nx.Graph()
        individuals = list(set(list(Data.interaction.i) + list(Data.interaction.j)))
        G.add_nodes_from(individuals)
        df_edge = Data.interaction.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
        for col, row in df_edge.iterrows():
            G.add_edge(row["i"], row["j"])

        return G

    else:

        if label is None:
            if replace_attr:
                Data.replace_str_attr_to_float()

            G = nx.Graph()
            G.add_nodes_from(Data.metadata["ID"].values)
            df_edge = Data.interaction.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])

            attribute_list = list(Data.metadata.columns.values)
            attr_dict = {}
            for index, row in Data.metadata.iterrows():
                id_dict = {}
                for i in attribute_list[1:]:
                    id_dict[str(i)] = row[str(i)]
                attr_dict[row[attribute_list[0]]] = id_dict

            nx.set_node_attributes(G, attr_dict)
            return G

        else:
            metadata = Data.metadata
            metadata = metadata.fillna("NaN")
            metadata = metadata.loc[metadata[label] != "NaN"]
            if replace_attr:
                metadata = replace_str_attr_to_float_2(metadata)

            metadata = metadata[["ID", label]]

            df_merge_data_to_filter = Data.interaction.merge(metadata, left_on=["i"], right_on=["ID"], how="inner")
            interaction = df_merge_data_to_filter.merge(metadata, left_on=["j"], right_on=["ID"], how="inner")
            interaction = interaction[["Time", "i", "j"]]

            G = nx.Graph()
            G.add_nodes_from(metadata["ID"].values)
            df_edge = interaction.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])

            attribute_list = list(metadata.columns.values)
            attr_dict = {}
            for index, row in metadata.iterrows():
                id_dict = {}
                for i in attribute_list[1:]:
                    id_dict[str(i)] = row[str(i)]
                attr_dict[row[attribute_list[0]]] = id_dict

            nx.set_node_attributes(G, attr_dict)
            return G


def hopping_time_networks(Data, minutes=1000):
    r"""Create multiple Networkx Graphs and DataFrames

        Creating multiple Networkx Graphs and DataFrames based on the given hopping time interval.

        Parameters
        ----------
        Data: Data
            Data Object that contains Tij- and Metadata for a data set.
        minutes: int
            The interval time in which the Networkx Graphs should be splitted.

        Returns
        -------
        network_list: list
            A list of all networkx Graphs for the given interval.
        df_list: list
             A list of all dataframes for the given interval.

        Examples
        --------
        In this example the full dataframe got splitted in dataframes and networks with time windows of 40 seconds
        (2/3 minutes). The output in this case describes the network and the dataframe for the first 40 seconds in the
        original dataframe.
        
        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta = "\\t",
        >>>                meta_attr_list = attr_list)
        >>> test_network_list, test_df_list = hopping_time_networks(test_df, minutes=2/3)
        >>> print(test_network_list[0].edges)
        [(0, 1), (1, 2), (1, 3), (2, 3)]
        >>> print(test_network_list[0].nodes)
        [0, 1, 2, 3]
        >>> for i in test_network_list[0].nodes:
            >>> print(test_network_list[0].nodes[i])
        {'Age': 1.0, 'Sex': 'F'}
        {'Age': 0.0, 'Sex': nan}
        {'Age': nan, 'Sex': 'M'}
        {'Age': 0.0, 'Sex': 'F'}
        >>> print(test_df_list[0])
           Time  i  j  TimeGroup
        0    20  0  1   0.0-40.0
        1    40  1  2   0.0-40.0
        2    40  1  3   0.0-40.0
        3    40  2  3   0.0-40.0

        See Also
        ---------
        face2face.imports.create_network.create_sliding_time_networks
        face2face.imports.create_network.event_time_networks

        """
    interval_seconds = minutes * 60

    time_stamps = Data.interaction["Time"][0]
    end_time_stamp = Data.interaction["Time"].tail(1).item()

    time_stamp_split_list = []

    id_group = 0
    while time_stamps <= end_time_stamp + interval_seconds:
        time_stamp_split_list.append([time_stamps, str(time_stamps-interval_seconds-Data.interaction["Time"][0]) + "-"
                                      + str(time_stamps-Data.interaction["Time"][0])])
        time_stamps += interval_seconds
        id_group += 1

    marker_timestamp_list = []
    unique_timestamp_list = set(Data.interaction["Time"].values)
    for i in range(1, len(time_stamp_split_list)):
        for j in unique_timestamp_list:
            if time_stamp_split_list[i][0] > j >= time_stamp_split_list[i-1][0]:
                marker_timestamp_list.append([j, time_stamp_split_list[i][0], time_stamp_split_list[i][1]])
    timestamp_marker_df = pd.DataFrame(marker_timestamp_list, columns=["Time", "sortby", "TimeGroup"])
    df_marker = Data.interaction.merge(timestamp_marker_df, how="left", on="Time")
    df_marker["sortby"] = df_marker["sortby"] - Data.interaction["Time"][0]
    df_list = [x for _, x in df_marker.groupby("sortby")]
    df_list = [i.drop(columns="sortby") for i in df_list]

    if hasattr(Data, "metadata") is False:
        network_list = []
        for i in df_list:
            G = nx.Graph()
            node_list = list(set(i["i"].append(i["j"])))
            G.add_nodes_from(node_list)
            df_edge = i.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])
            network_list.append(G)

        return network_list, df_list

    else:
        network_list = []
        for i in df_list:
            G = nx.Graph()
            node_list = list(set(i["i"].append(i["j"])))
            G.add_nodes_from(node_list)
            df_edge = i.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])

            attribute_list = list(Data.metadata.columns.values)
            attr_dict = {}
            for index, row in Data.metadata.iterrows():
                id_dict = {}
                for i in attribute_list[1:]:
                    id_dict[str(i)] = row[str(i)]
                attr_dict[row[attribute_list[0]]] = id_dict

            nx.set_node_attributes(G, attr_dict)
            network_list.append(G)
        return network_list, df_list


def sliding_time_networks(Data, slide=1, interval=1000):
    r"""Create multiple Networkx Graphs and DataFrames

        Creating multiple Networkx Graphs and DataFrames based on the given interval and the sliding time interval.

        Parameters
        ----------
        Data: Data
            Data Object that contains Tij- and Metadata for a data set.
        slide: int
            The time steps in which the intervals should be created.
        interval: int
            The interval time in which the networkx Graphs should be splitted.

        Returns
        -------
        network_list: list
            A list of all networkx Graphs for a given interval.
        df_list: list
            A list of all dataframes for a given interval.

        Examples
        ---------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta = "\t",
        >>>                meta_attr_list=attr_list)
        >>> test_network_list, test_df_list = sliding_time_networks(test_df, slide=1/3, interval=2/3)
        >>> print(test_df_list[0])
           Time  i  j  TimeGroup
        0    20  0  1     0-40.0
        1    40  1  2     0-40.0
        3    40  1  3     0-40.0
        5    40  2  3     0-40.0
        >>> print(test_network_list[0].nodes)
        [0, 1, 2, 3]
        >>> print(test_network_list[0].edges)
        [(0, 1), (1, 2), (1, 3), (2, 3)]
        >>> print(test_df_list[1])
            Time  i  j  TimeGroup
        2     40  1  2  20.0-60.0
        4     40  1  3  20.0-60.0
        6     40  2  3  20.0-60.0
        7     60  4  6  20.0-60.0
        9     60  4  7  20.0-60.0
        11    60  6  7  20.0-60.0
        >>> print(test_network_list[1].nodes)
        [1, 2, 3, 4, 6, 7]
        >>> print(test_network_list[1].edges)
        [(1, 2), (1, 3), (2, 3), (4, 6), (4, 7), (6, 7)]

        See Also
        ---------
        face2face.imports.create_network.hopping_time_networks
        face2face.imports.create_network.event_time_networks
        """
    interval_seconds = interval * 60
    interval_slide = slide * 60

    time_stamps = Data.interaction["Time"][0]
    end_time_stamp = Data.interaction["Time"].tail(1).item()
    time_stamp_split_list = []

    interval = 0
    interval2 = interval_seconds
    id_group = 0
    while time_stamps <= end_time_stamp + interval_seconds:
        time_stamp_split_list.append([time_stamps, time_stamps + interval_seconds, id_group, str(interval) + "-"
                                      + str(interval2)])
        time_stamps += interval_slide
        interval2 += interval_slide
        interval += interval_slide
        id_group += 1

    marker_timestamp_list = []
    unique_timestamp_list = set(Data.interaction["Time"].values)
    for i in time_stamp_split_list:
        for j in unique_timestamp_list:
            if i[1] > j >= i[0]:
                marker_timestamp_list.append([j, i[2], i[3]])

    timestamp_marker_df = pd.DataFrame(marker_timestamp_list, columns=["Time", "id_group","TimeGroup"])
    df_marker = Data.interaction.merge(timestamp_marker_df, how="left", on="Time")

    df_list = [x for _, x in df_marker.groupby("id_group")]
    df_list = [i.drop(columns="id_group") for i in df_list]

    if hasattr(Data, "metadata") is False:
        network_list = []
        for i in df_list:
            G = nx.Graph()
            node_list = list(set(i["i"].append(i["j"])))
            G.add_nodes_from(node_list)
            df_edge = i.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])
            network_list.append(G)

        return network_list, df_list

    else:

        network_list = []
        for i in df_list:
            G = nx.Graph()
            node_list = list(set(i["i"].append(i["j"])))
            G.add_nodes_from(node_list)
            df_edge = i.groupby(["i", "j"]).size().reset_index(name="EdgeWeight")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])

            attribute_list = list(Data.metadata.columns.values)
            attr_dict = {}
            for index, row in Data.metadata.iterrows():
                id_dict = {}
                for i in attribute_list[1:]:
                    id_dict[str(i)] = row[str(i)]
                attr_dict[row[attribute_list[0]]] = id_dict

            nx.set_node_attributes(G, attr_dict)
            network_list.append(G)
        return network_list, df_list


def event_time_networks(Data, events):
    r"""Create multiple Networkx Graphs and DataFrames

        Creating multiple Networkx Graphs and DataFrames based on a given event list.

        Parameters
        ----------
        Data: Data
            Data Object that contains Tij- and Metadata for a data set.
        events: list
            A list of tuples that contains the eventname and the start and the beginning of the event.

        Returns
        -------
        network_list: list
            A list of all networkx Graphs for a given interval.
        df_list: list
            A list of all dataframes for a given interval.

        Examples
        --------
        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>       path_meta="face2face/data/Test/meta_test.dat", separator_meta = "\t",
        >>>       meta_attr_list=attr_list)

        >>> event_list = [("Event A", 20, 60), ("Event B", 120, 180), ("Event C", 400, 500)]

        >>> test_network_list, test_df_list = event_time_networks(test_df, events=event_list)
        >>> print(test_df_list[0])
             Time i  j TimeGroup
        1    40  1  2   Event A
        2    40  1  3   Event A
        3    40  2  3   Event A
        4    60  4  6   Event A
        5    60  4  7   Event A
        6    60  6  7   Event A
        >>> print(test_network_list[0].nodes)
        [1, 2, 3, 4, 6, 7]
        >>> print(test_network_list[0].edges)
        [(1, 2), (1, 3), (2, 3), (4, 6), (4, 7), (6, 7)]
        >>> print(test_df_list[1])
             Time i  j TimeGroup
        13   140  7  6   Event B
        14   140  4  6   Event B
        15   140  7  4   Event B
        16   160  7  6   Event B
        17   160  4  6   Event B
        18   160  7  4   Event B
        19   180  7  6   Event B
        20   180  4  6   Event B
        21   180  7  4   Event B
        >>> print(test_network_list[1].nodes)
        [4, 6, 7]
        >>> print(test_network_list[1].edges)
        [(4, 6), (4, 7), (6, 7)]
        >>> print(test_df_list[2])
             Time  i   j   TimeGroup
        25   420   4   6   Event C
        26   420   4   7   Event C
        27   420   6   7   Event C
        28   440   4   6   Event C
        29   440   4   7   Event C
        30   440   6   7   Event C
        31   440   1   2   Event C
        32   440   1   3   Event C
        33   440   2   3   Event C
        34   460   1   2   Event C
        35   460   1   3   Event C
        36   460   2   3   Event C
        37   460  10  11   Event C
        38   460  11  12   Event C
        39   460  10  12   Event C
        40   500  10  11   Event C
        41   500  11  12   Event C
        42   500  10  12   Event C
        >>> print(test_network_list[2].nodes)
        [1, 2, 3, 4, 6, 7, 10, 11, 12]
        >>> print(test_network_list[2].edges)
        [(1, 2), (1, 3), (2, 3), (4, 6), (4, 7), (6, 7), (10, 11), (10, 12), (11, 12)]

        See Also
        --------
        face2face.imports.create_network.hopping_time_networks
        face2face.imports.create_network.sliding_time_networks

        """

    marker_timestamp_df = []
    unique_timestamp_list = set(Data.interaction["Time"].values)
    for i in range(len(events)):
        for j in unique_timestamp_list:
            if events[i][2] >= j > events[i][1]:
                marker_timestamp_df.append([j, events[i][0], i])
    marker_timestamp_df = pd.DataFrame(marker_timestamp_df, columns=["Time","Event","TimeGroup"])
    df_marker = Data.interaction.merge(marker_timestamp_df, how="left", on="Time")

    df_list = [x for _, x in df_marker.groupby("TimeGroup")]
    df_list = [i.drop(columns="TimeGroup") for i in df_list]

    if hasattr(Data, "metadata") is False:
        network_list = []
        for i in df_list:
            G = nx.Graph()
            node_list = list(set(i["i"].append(i["j"])))
            G.add_nodes_from(node_list)
            df_edge = i.groupby(["i", "j"]).size().reset_index(name="AmountOfContactsIJ")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])
            network_list.append(G)

        return network_list, df_list

    else:
        network_list = []
        for i in df_list:
            G = nx.Graph()
            node_list = list(set(i["i"].append(i["j"])))
            G.add_nodes_from(node_list)
            df_edge = i.groupby(["i", "j"]).size().reset_index(name="EdgeWeight")
            for col, row in df_edge.iterrows():
                G.add_edge(row["i"], row["j"])

            attribute_list = list(Data.metadata.columns.values)
            attr_dict = {}
            for index, row in Data.metadata.iterrows():
                id_dict = {}
                for i in attribute_list[1:]:
                    id_dict[str(i)] = row[str(i)]
                attr_dict[row[attribute_list[0]]] = id_dict

            nx.set_node_attributes(G, attr_dict)
            network_list.append(G)
        return network_list, df_list


def replace_str_attr_to_float_2(metadata):

    r"""Replaces string attributes

        If replaced_attr parameter is set true in the `create_network_from_data` function, string attributes will be
        replaced by float numbers.

        Parameters
        ----------
        metadata : DataFrame
            DataFrame object that contains meta data for the persons that have participated in the experiment.

        Returns
        -------
        metadata : DataFrame
            DataFrame object that contains meta data for the persons that have participated in the experiment. Metadata
            string attribute values are changed into float values.
    """
    parameter_list = []
    for col in metadata.columns:
        if col != "ID":
            parameter_list.append(col)

    parameter_specification_list = []
    for i in parameter_list:
        contacts_of_i = list(set(metadata[i]))
        parameter_specification_list.append(contacts_of_i)

    mapping = {}
    for i in parameter_specification_list:
        for j in range(len(i)):
            if isinstance(i[j], str):
                mapping[i[j]] = float(j)
    metadata = metadata.replace(mapping)
    return metadata



import powerlaw


def search_best_fit_distribution(data, xmin=None, xmax=None, extend_data=None, discret=False, approximation_method=None,
                                 distance_measurement_method="D", sigma_threshold=None, parameter_range=None,
                                 termination_condition=1):

    """Searches the best fitting distribution

        Creates a fit object and compares every distribution for the given data set to find the best fitting one.

        Parameters
        ----------
        data: list
            A list of all the $\delta t$ values for the analysed interaction duration.
        xmin: float, default: None
            Set xmin to localize the part of the data set that you want to analyze.
        xmax: float, default: None
            Set xmax to localize the port of the data set that you you want to analyze.
        extend_data: int, default: None
            Can be used to extend the data set with random generated data points.
        discret: bool, default: False
            Must be set if data set is discrete, because discrete versions of probability distributions are calculated
            differently.
        approximation_method: str, int
            Different approximation methods for discrete data sets. Possible values: "round", "xmax" or an integer.
        distance_measurement_method: str, default: "D".
            Changes the measurement method to calculated the minimized distance between the empirical data and the
            distribution. Possible Values: "D","V","Asquare".
        sigma_threshold: float, default: None
            Can be set to add a constraint for the maximum sigma.
        parameter_range: dict, default: None
            Can be set to limit multiple parameters for the distributions.
        termination_condition: int, default: 1
            If the function get_best_fit_distribution does not find best fit, because the remaining two are too similar
            you can set it on 2, to get the two best fitting distributions.

        Returns
        -------
        Fit: Fit Object
            Contains the created Fit Object to use it for plotting as PDF, CDF or CCDF.
        tupel_list: list
            Contains tupels with the comparison of every distribution that is included in the package.
        best_fit: list
            Contains the best fitting distributions as strings. Amount of best fitting distributions depends on
            the termination_condition parameter.


        See Also
        ---------
        face2face.statistics.distribution.calculate_contact_duration
        face2face.statistics.distribution.calculate_triangle_duration
        face2face.statistics.distribution.calculate_inter_contact_duration

        """

    fit = powerlaw.Fit(data, xmin=xmin, xmax=xmax, discret=discret, discrete_approximation=approximation_method,
                       sigma_threshold=sigma_threshold, parameter_range=parameter_range
                       , xmin_distance=distance_measurement_method)
    if extend_data is not None:
        fit = fit.power_law.generate_random(extend_data)

    if sigma_threshold is not None:
        if fit.noise_flag:
            print("Sigma Treshold was too small to get a result")
        else:
            print("Sigma Treshold was applied successfully")

    distribution_list = ["power_law", "lognormal", "exponential", "truncated_power_law", "stretched_exponential",
                         "lognormal_positive"]

    best_fit = get_best_fit_distribution(distribution_list, fit, termination_condition=termination_condition)
    if len(best_fit) == 1:
        print("The best fitting distribution is " + best_fit[0])
    elif len(best_fit) == 2:
        print("The best fitting distribution is " + best_fit[0] + "or" + best_fit[1])

    tupel_list = distribution_tupel_combinations(distribution_list, fit)

    return fit, tupel_list, best_fit


def get_best_fit_distribution(distribution_list, fit, termination_condition=1):

    """Searches the best fitting distribution

        Tests the included distributions recursively against each other to get the best fitting distribution.

        Parameters
        ----------
        distribution_list: list
            Contains a list of strings with every distribution included in the powerlaw library.
        fit: Fit object
            Fit objected for the given data set.
        termination_condition: int, default: 1
            If the function get_best_fit_distribution does not find best fit, because the remaining two are too similar
            you can set it on 2, to get the two best fitting distributions.

        Returns
        -------
        distribution_list: list
            Contains the best fit distribution as a string. Amount of best fitting distributions depends on
            the termination_condition parameter.

        See Also
        ---------
        face2face.statistics.check_distribution.search_best_fit_distribution
        """

    if len(distribution_list) == termination_condition:
        return distribution_list

    else:
        distribution_list2 = []
        for i in range(1, len(distribution_list)):
            r, p = fit.distribution_compare(distribution_list[i], distribution_list[i - 1], normalized_ratio=True)
            if r > 0:
                if distribution_list[i] not in distribution_list2:
                    distribution_list2.append(distribution_list[i])
            elif r < 0:
                if distribution_list[i - 1] not in distribution_list2:
                    distribution_list2.append(distribution_list[i - 1])
            else:
                if distribution_list[i] not in distribution_list2:
                    distribution_list2.append(distribution_list[i])
                if distribution_list[i - 1] not in distribution_list2:
                    distribution_list2.append(distribution_list[i - 1])
        return get_best_fit_distribution(distribution_list2, fit)


def distribution_tupel_combinations(distribution_list, fit):
    """Searches the best fitting distribution

        Compares every distribution included in this package against each other and returns every combination as a list
        of tupels. The normalized loglikelihood is also included in the tupel.

        Parameters
        ----------
        distribution_list: list
            Contains a list of strings with every distribution included in this package.
        fit: Fit object
            Fit objected for the given data set.

        Returns
        -------
        list_for_distribution_tupels: list
            Contains tupels with the comparison of every distribution that is included in the package.

        See Also
        ---------
        face2face.statistics.check_distribution.search_best_fit_distribution
        """
    list_for_distribution_tupels = []
    for i in range(len(distribution_list)):
        for j in range(len(distribution_list)):
            if i != j:
                r, p = fit.distribution_compare(distribution_list[i], distribution_list[j], normalized_ratio=True)
                if r > 0:
                    if (r, distribution_list[i], distribution_list[j]) or \
                            (r, distribution_list[j], distribution_list[i]) not in list_for_distribution_tupels:
                        distribution_tupel = (r, distribution_list[i], distribution_list[j])
                        list_for_distribution_tupels.append(distribution_tupel)

    return list_for_distribution_tupels





from .visualization.plot_probability_distribution import plot_contact_duration
from .visualization.plot_probability_distribution import plot_triangle_duration
from .visualization.plot_probability_distribution import plot_inter_contact_duration
from .visualization.plot_average_degree_group import plot_avg_degree_group
from .visualization.plot_contact_matrix_heatmap import plot_cm_heatmap
from .visualization.output_network_information import print_network_measures
from .visualization.plot_histogram_null_model import *

from .statistics.distribution import calculate_triangle_duration, calculate_inter_contact_duration, \
    calculate_contact_duration

from .statistics.average_degree import avg_degree_attr, group_list_degree, global_avg_var_std

from .statistics.null_modell import shuffle_label_z_score_mixing_matrix, mapping_function, to_simple_graph,\
    remove_self_loops,configuration_model_label_z_score_mixing_matrix, check_bonferroni_correction

from .statistics.network_quantities import mean_degree_network, variance_std_network, \
    calculating_number_of_edges_nodes, average_path_length_network, clustering_coefficient

from .statistics.check_distribution import search_best_fit_distribution, get_best_fit_distribution, distribution_tupel_combinations

from .imports.load_all_data import Data

from .imports.create_network import create_network_from_data, sliding_time_networks, hopping_time_networks, \
    event_time_networks

from .compatibility_check.check_compatibility import check_compatibility_of_versions, check_which_methods_are_affected



import math

import pandas as pd

from face2face.imports.create_network import create_network_from_data
from face2face.imports.load_all_data import Data

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)


def avg_degree_attr(df):
    """Calculates the average degree

        Calculates the average degree for every subgroup and for the whole group for every attribute.

        Parameters
        ----------
        df: Data
            Data Object that contains Tij- and Metadata for a data set.

        Returns
        -------
        attr_degree_list : list
            A list that contains the average degree for every subgroup and for the whole group for every attribute.

        Examples
        ---------
        The first string in a list, for example "Age", tells you which attribute's average degrees are in this list.
        The following first list entries for example 0.0, 1.0 or 2.0 are the different attribute values and the second
        entry in the list is the average degree for this attribute value. The 'GlobalAvG' list entry tells you the
        average degree for the whole attribute (for example "Age").

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>               meta_attr_list=attr_list)
        >>> avg_degree_list = avg_degree_attr(test_df)
        >>> print(avg_degree_list)
        [['Age', [[0.0, 1.6666666666666667], [1.0, 2.5], [2.0, 2.0], ['GlobalAvG', 2.055555555555556]]],
         ['Sex', [['F', 1.8], ['M', 2.0], ['GlobalAvG', 1.9]]]]

        See Also
        ---------
        face2face.group_list_degree
        face2face.global_avg_var_std

        """

    network = create_network_from_data(df)
    df_meta_nan = df.metadata.fillna("NaN")

    parameter_list = []
    for col in df_meta_nan.columns:
        if col != "ID":
            parameter_list.append(col)

    complete_parameter_value_list = []
    for i in parameter_list:
        parameter_value_list = []
        nan_filtered_dataframe = df.metadata.loc[df.metadata[i] != "NaN"]
        for parameter_values, grouped_by_dataframes in nan_filtered_dataframe.groupby(i):
            parameter_value_list.append([grouped_by_dataframes["ID"], parameter_values])
        complete_parameter_value_list.append([i, parameter_value_list])

    avg_degree_param_list = []
    for i in complete_parameter_value_list:
        value_avg_degree_pair_list = []
        for j in i[1]:
            avg_degree = 0
            for k in j[0]:
                avg_degree += network.degree[k]
            avg_degree = avg_degree / len(j[0])
            value_avg_degree_pair_list.append([j[1], avg_degree])
        avg_degree_param_list.append([i[0], value_avg_degree_pair_list])

    for i in avg_degree_param_list:
        avg_degree_parameter = 0
        for j in i[1]:
            avg_degree_parameter += j[1]
        avg_degree_parameter = avg_degree_parameter / len(i[1])
        i[1].append(["GlobalAvG", avg_degree_parameter])

    return avg_degree_param_list


def group_list_degree(df):
    """Creates lists of the degrees

       Creating a list of degrees for every attribute value of every group

       Parameters
       ----------
       df: Data
           Contains a dataframe with the tij-data from the data set

       Returns
       -------
       attr_degree_list: list
           Contains lists for all attributes and all occurring degrees for the different attribute values.

       Examples
       ---------
       This functions return a list of lists which contains a list for every attribute value. In this list you have the
       attribute as a string, the attribute value and a list with all degrees for this attribute value.

       >>> attr_list = ["ID", "Age", "Sex"]
       >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
       >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
       >>>               meta_attr_list=attr_list)
       >>> attr_degree_list = group_list_degree(test_df)
       [['Age', 1.0, [1, 2, 2]], ['Age', 0.0, [3, 2]], ['Age', 2.0, [2, 2, 2]], ['Sex', 'F', [1, 2, 2, 2, 2]],
       ['Sex', 'M', [2, 2, 2]]]


       See Also
       ----------
       face2face.avg_degree_attr
       face2face.global_avg_var_std

       """
    df_meta_nan = df.metadata.fillna("NaN")
    network = create_network_from_data(df)

    parameter_list = []
    for col in df_meta_nan.columns:
        if col != "ID":
            parameter_list.append(col)

    complete_parameter_value_list = []
    for i in parameter_list:
        nan_filtered_dataframe = df.metadata.loc[df.metadata[i] != "NaN"]
        for parameter_values, grouped_by_dataframes in nan_filtered_dataframe.groupby(i):
            complete_parameter_value_list.append([i, parameter_values, list(grouped_by_dataframes["ID"])])

    for i in complete_parameter_value_list:
        parameter_value_degree_list = []
        for j in i[2]:
            parameter_value_degree_list.append(network.degree(j))
        i[2] = parameter_value_degree_list[:]
    return complete_parameter_value_list


def global_avg_var_std(attr_degree_list):
    """Calculating global degree measures

    Calculate the average, the variance and the standard deviation for the degrees of the whole data set.

    Parameters
    ----------
    attr_degree_list : list
        This list is the output of :mod:`~face2face.statistics.average_degree.avg_degree_attribute`.

    Returns
    -------
    measures : list
        This list contains the average degree, the variance and the standard deviation for the degrees of the whole
        data set.

    Examples
    --------
    You can use the output of the :meth:`avg_degree_attr` method as input for this function. The return of this function
    is the average degree, the variance of the degree and the standard deviation for the degree of all attributes.

    >>> attr_degree_list = avg_degree_attr(test_df)
    >>> print(attr_degree_list)
    [['Age', 1.0, [1, 2, 2]], ['Age', 0.0, [3, 2]], ['Age', 2.0, [2, 2, 2]], ['Sex', 'F', [1, 2, 2, 2, 2]],
    ['Sex', 'M', [2, 2, 2]]]
    >>> measures = global_avg_var_std(attr_degree_list)
    >>> print(measures)
    [1.9777777777777779, 0.006049382716049409, 0.07777777777777795]

    See Also
    ---------
    face2face.avg_degree_attr
    face2face.group_list_degree

    """
    total_sum = 0
    for i in attr_degree_list:
        total_sum += i[1][-1][1]
    global_mean = total_sum/len(attr_degree_list)

    variance = 0
    for i in attr_degree_list:
        variance += (i[1][-1][1] - global_mean)**2
    variance = variance/len(attr_degree_list)

    std = math.sqrt(variance)
    measures = [global_mean, variance, std]
    return measures




import pandas as pd


pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)
pd.set_option('display.float_format', lambda x: '%.5f' % x)


class Data:

    def __init__(self, data_set_name=None, path_tij=None, separator_tij=None, path_meta=None, separator_meta=None,
                 meta_attr_list=None, header=None, meta_df=None, tij_df=None):

        """Initialize an object that contains Tij- and Metadata for imported or given data sets.

            Creating multiple Networkx Graphs and DataFrames based on a given event list.

            Parameters
            ----------
            data_set_name: optional: str, default: None
                If you want to use predefined data sets you can use this parameter to create a Data object. For example
                you can import the "Synthetic" data set.
            path_tij: optional: str, default: None
                Must be used to import tij data sets manually.
            separator_tij: optional: str, default: None
                Must be used if you import tij data sets manually.
            path_meta: optional: str, default: None
                Must be used to import meta data sets manually.
            separator_meta: optional: str, default: None
                 Must be used if you import meta data sets manually.
            meta_attr_list: optional: list, default: None
                Must be used if you want to import metadata manually that has no header.
            header: optional: str, default: None
                Must be used if tij and/or metadata already contain a header. Possible parameters are 'tij', 'meta' or
                'all'.
            meta_df: optional: dataframe, default: None
                Must be used if you want to create a Data Class Object based on a dataframe
            tij_df: optional: dataframe, default: None
                Must be used if you want to create a Data Class Object based on a dataframe, for example with a
                dataframe of the dataframe list from the create_network functions

            Examples
            ---------
            Importing predefined data sets by their string name

            >>> df = Data("Synthetic")

            Importing own data sets or predefined data sets manually (without header)

            >>> attr_list = ["ID", "Age", "Sex"]
            >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\\t",
            >>>       path_meta="face2face/data/Test/meta_test.dat", separator_meta = "\\t",
            >>>       meta_attr_list=attr_list)

            Importing own data sets or predefined data sets manually (with headers for tij- and metadata)

            >>> test_df = Data(path_tij="face2face/data/Synthetic_Data/synthetic_tij.dat", separator_tij=",",
            >>>       path_meta="face2face/data/Synthetic_Data/synthetic_metadata.dat", separator_meta = ",", header= "all")

            """
        script_path = __file__
        script_path = script_path.replace("\\", "/")
        path = str(script_path.replace("imports/load_all_data.py", ""))

        data_set = {"test": {"tij": [str(path + "data/Test/tij_test.dat"), "\t"],
                             "meta": [str(path + "data/Test/meta_test.dat"), "\t"],
                             "attr": {0: "ID", 1: "Age", 2: "Gender"}},
                    "Synthetic": {"tij": [str(path + "data/Synthetic_Data/synthetic_tij.dat"), ","],
                                  "meta": [str(path + "data/Synthetic_Data/synthetic_metadata.dat"), ","],
                                  "attr": {0: "ID", 1: "type"}},
                    "WS16": {"tij": [str(path + "data/WS16/tij_WS16.dat"), "\t"],
                             "meta": [str(path + "data/WS16/metadata_WS16.dat"), "\t"],
                             "attr": {0: "ID", 1: "Age", 2: "Sex", 3: "Country", 4: "Language", 5: "Education",
                                      6: "Academic Background", 7: "Role", 8: "Previous participation"}},
                    "ICCSS17": {"tij": [str(path + "data/ICCSS17/tij_ICCSS17.dat"), "\t"],
                             "meta": [str(path + "data/ICCSS17/metadata_ICCSS17.dat"), "\t"],
                             "attr": {0: "ID", 1: "Age", 2: "Sex", 3: "Country", 4: "Language", 5: "Education",
                                      6: "Academic Background", 7: "Role", 8: "Previous participation"}},
                    }

        if data_set_name is not None:
            if len(data_set[data_set_name]) == 3:
                self.interaction = pd.read_csv(data_set[data_set_name]["tij"][0], sep=data_set[data_set_name]["tij"][1],
                                               header=None, engine='python', index_col=False)
                self.metadata = pd.read_csv(data_set[data_set_name]["meta"][0], sep=data_set[data_set_name]["meta"][1],
                                            header=None, engine='python', index_col=False)

                self.interaction = self.interaction.rename(columns={0: "Time", 1: "i", 2: "j"})
                self.interaction.index.name = 'Index'
                self.metadata = self.metadata.rename(columns=data_set[data_set_name]["attr"])
                self.metadata.index.name = 'Index'
            elif len(data_set[data_set_name]) == 1:
                self.interaction = pd.read_csv(data_set[data_set_name]["tij"][0], sep=data_set[data_set_name]["tij"][1],
                                               header=None, engine='python', index_col=False)

                self.interaction = self.interaction.rename(columns={0: "Time", 1: "i", 2: "j"})
                self.interaction.index.name = 'Index'

        elif data_set_name is None and header is None:
            if meta_attr_list is not None:
                self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=None, engine='python',
                                               index_col=False)
                self.metadata = pd.read_csv(path_meta, sep=separator_meta, header=None, engine='python',
                                            index_col=False)

                self.interaction = self.interaction.rename(columns={0: "Time", 1: "i", 2: "j"})
                self.interaction.index.name = 'Index'
                columns = {}
                for i in range(len(meta_attr_list)):
                    columns[i] = meta_attr_list[i]
                self.metadata = self.metadata.rename(columns=columns)
                self.metadata.index.name = 'Index'
            elif meta_attr_list is None and path_tij is not None:

                self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=None, engine='python',
                                               index_col=False)
                self.interaction = self.interaction.rename(columns={0: "Time", 1: "i", 2: "j"})
                self.interaction.index.name = 'Index'

        elif data_set_name is None and header is not None:
            if header == "meta":

                self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=None, engine='python',
                                               index_col=False)
                self.metadata = pd.read_csv(path_meta, sep=separator_meta, header=0, engine='python',
                                            index_col=False)
                self.interaction = self.interaction.rename(columns={0: "Time", 1: "i", 2: "j"})
                self.interaction.index.name = 'Index'
                self.metadata.index.name = 'Index'

            elif header == "tij":
                if meta_attr_list is None and path_meta is None:

                    self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=0, engine='python',
                                                   index_col=False)
                    self.interaction.index.name = 'Index'
                else:
                    if path_tij and path_meta and meta_attr_list is not None:

                        self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=0, engine='python',
                                                       index_col=False)
                        self.metadata = pd.read_csv(path_meta, sep=separator_meta, header=None, engine='python',
                                                    index_col=False)

                        self.interaction.index.name = 'Index'
                        columns = {}
                        for i in range(len(meta_attr_list)):
                            columns[i] = meta_attr_list[i]
                        self.metadata = self.metadata.rename(columns=columns)
                        self.metadata.index.name = 'Index'

            elif header == "all":
                if path_tij and path_meta is not None:

                    self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=0, engine='python',
                                                   index_col=False)
                    self.metadata = pd.read_csv(path_meta, sep=separator_meta, header=0, engine='python',
                                                index_col=False)

                    self.interaction.index.name = 'Index'
                    self.metadata.index.name = 'Index'
                else:
                    self.interaction = pd.read_csv(path_tij, sep=separator_tij, header=0, engine='python',
                                                   index_col=False)
                    self.interaction.index.name = 'Index'

        if meta_df is not None and tij_df is not None:
            self.interaction = tij_df
            self.metadata = meta_df
            self.interaction.index.name = 'Index'
            self.metadata.index.name = 'Index'
        elif meta_df is None and tij_df is not None:
            self.interaction = tij_df
            self.interaction.index.name = 'Index'

    def replace_str_attr_to_float(self):

        """Replace string attribute values

        Replaces string attributes into float attributes for the null model functions.

        Parameters
        ----------
        self : Data
            Data Object that contains Tij- and Metadata for a data set.

        Returns
        --------
        self : Data
            Data Object that contains Tij- and Metadata for a data set. Metadata string attribute values are changed
            into float values.

        Examples
        ---------
        As you can see in this example every non-numerical or NaN value gets replaced into a float value. In this case
        "F" turns into 0.0 and "M" turns into 1.0.

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>                meta_attr_list=attr_list)
        >>> print(test_df.metadata)
                ID     Age  Sex
        Index
        0       0 1.00000    F
        1       1 0.00000  NaN
        2       2     nan    M
        3       3 0.00000    F
        4       4     nan    M
        5       6 1.00000    F
        6       7 1.00000  NaN
        7      10 2.00000    F
        8      11 2.00000    M
        9      12 2.00000    F
        >>> test_df.replace_str_attr_to_float()
        >>> print(test_df.metadata)
                ID     Age    Sex
        Index
        0       0 1.00000 0.00000
        1       1 0.00000     NaN
        2       2     NaN 1.00000
        3       3 0.00000 0.00000
        4       4     NaN 1.00000
        5       6 1.00000 0.00000
        6       7 1.00000     NaN
        7      10 2.00000 0.00000
        8      11 2.00000 1.00000
        9      12 2.00000 0.00000

        """

        self.metadata = self.metadata.fillna("NaN")

        parameter_list = []
        for col in self.metadata.columns:
            if col != "ID":
                parameter_list.append(col)

        parameter_specification_list = []
        for i in parameter_list:
            contacts_of_i = list(set(self.metadata[i]))
            parameter_specification_list.append(contacts_of_i)

        mapping = {}
        for i in parameter_specification_list:
            for j in range(len(i)):
                if isinstance(i[j], str) and i[j] != "NaN":
                    mapping[i[j]] = float(j)
        self.metadata = self.metadata.replace(mapping)
        return self







import numpy as np


def calculate_contact_duration(Data, bins=None):
    """Calculates the probabilitys for the contact duration

    Calculates the probabilities for having a face-to-face contact for the different times delta T

    Parameters
    ----------
    Data: Data
        Contains a dataframe with the tij-data from the data set. Might also contain metadata.
    bins: optional: int, default: None
        Amount of bins that are used to calculate the probabilities based on a histogram.

    Returns
    -------
    x_y_list: List
        A list with two lists. The first list contains the probabilities for the y-axis and the second list contains
        the contact duration delta T
    delta_t_list: list
        Contains all the calculated $\delta t$ values for the contact duration of the used data set.

    References
    ----------
    .. [1] Cattuto C, Van den Broeck W, Barrat A, Colizza V, Pinton JF, et al. (2010) Dynamics
           of Person-to-Person Interactions from Distributed RFID Sensor Networks. PLOS ONE

    Examples
    ----------

    >>> attr_list = ["ID", "Age", "Sex"]
    >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
    >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
    >>>               meta_attr_list=attr_list)
    >>> y_x_list, delta_t_list = calculate_contact_duration(test_df)
    >>> print(y_x_List[0])
    [0.45454545454545453, 0.2727272727272727, 0.13636363636363635, 0.13636363636363635]
    >>> print(y_x_list[1])
    [20, 40, 60, 80]

    See Also
    ---------
    face2face.statistics.distribution.calculate_triangle_duration
    face2face.statistics.distribution.calculate_inter_contact_duration

    """

    tij_data_sorted = Data.interaction.sort_values(by=["i", "j"])

    tij_data_sorted["diff"] = tij_data_sorted.groupby(["i", "j"])["Time"].diff()

    marker_list = []
    for key, value in tij_data_sorted.iterrows():
        if value["diff"] > 20:
            marker_list.append(1)
        else:
            marker_list.append(0)

    tij_data_sorted["Marker"] = marker_list
    tij_data_sorted["Ind"] = tij_data_sorted["Marker"].cumsum()

    tij_data_sorted = tij_data_sorted.groupby(["Ind", "i", "j"]).size().reset_index(name="Number")

    delta_t_list = []
    for key, value in tij_data_sorted.iterrows():
        delta_t = value["Number"] * 20
        delta_t_list.append(delta_t)

    if bins is None:
        bins = 10**np.linspace(np.log10(min(delta_t_list)), np.log10(max(delta_t_list)), 50)
        x_delta_t, y_probability = np.histogram(delta_t_list, bins=bins, density=True)
    else:
        x_delta_t, y_probability = np.histogram(delta_t_list, bins=bins, density=True)

    y_probability = y_probability[:-1]
    x_y_list = [list(x_delta_t), list(y_probability)]

    return x_y_list, delta_t_list


def calculate_triangle_duration(Data, bins=None):
    r"""Calculates the probabilities for the contact duration between triangle constellations

        Calculates the probability how long a triangle constellation lasts.

        Parameters
        ----------
        Data.interaction: Data
            Contains a dataframe with the tij-data from the dataset

        Returns
        -------
        x_y_list: List
            A list with two lists. The first list contains the probabilitys for the y-axis and the second list contains
            the triangle duration delta T
        delta_t_list: list
            Contains all the calculated $\delta t$ values for the contact duration of the used data set.

        References
        ----------
        .. [1] Cattuto C, Van den Broeck W, Barrat A, Colizza V, Pinton JF, et al. (2010) Dynamics of Person-to-Person
               Interactions from Distributed RFID Sensor Networks. PLOS ONE

        Examples
        ----------
        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\\t",
        >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\\t",
        >>>               meta_attr_list=attr_list)
        >>> y_x_list, delta_t_list = calculate_triangle_duration(test_df)
        >>> print(y_x_list[0])
        [0.42857142857142855, 0.2857142857142857, 0.14285714285714285, 0.14285714285714285]
        >>> print(y_x_list[1])
        [20, 40, 60, 80]

        See Also
        ---------
        face2face.statistics.distribution.calculate_contact_duration
        face2face.statistics.distribution.calculate_inter_contact_duration

        """
    df_merge_1 = Data.interaction.merge(Data.interaction, left_on=["Time", "i"], right_on=["Time", "j"])
    df_merge_2 = df_merge_1.merge(Data.interaction, left_on=["Time"], right_on=["Time"])

    df_filter_triangle = df_merge_2[(df_merge_2["i_x"] == df_merge_2["j_y"])
                                    & (df_merge_2["j_x"] == df_merge_2["j"])
                                    & (df_merge_2["i_y"] == df_merge_2["i"])]

    df_no_duplicates = df_filter_triangle.drop_duplicates(keep="first")

    df_no_duplicates["Diff"] = df_no_duplicates.groupby(["i_x", "j_x", "i_y"])["Time"].diff()

    marker_list = []
    for key, value in df_no_duplicates.iterrows():
        if value["Diff"] == 20:
            marker_list.append(0)
        elif value["Diff"] > 20:
            marker_list.append(1)
        else:
            marker_list.append(0)
    df_no_duplicates["Marker"] = marker_list
    df_no_duplicates["Ind"] = df_no_duplicates["Marker"].cumsum()

    df_merge_dd_gb = df_no_duplicates.groupby(["Ind", "i_x", "j_x", "i_y"]) \
        .size().reset_index(name="Number")

    delta_t_list = []
    for key, value in df_merge_dd_gb.iterrows():
        if value["Number"] == 1:
            delta_t = 20
            delta_t_list.append(delta_t)
        elif value["Number"] > 1:
            delta_t = value["Number"] * 20
            delta_t_list.append(delta_t)

    if bins is None:
        bins = 10**np.linspace(np.log10(min(delta_t_list)), np.log10(max(delta_t_list)), 50)
        x_delta_t, y_probability = np.histogram(delta_t_list, bins=bins, density=True)
    else:
        x_delta_t, y_probability = np.histogram(delta_t_list, bins=bins, density=True)

    y_probability = y_probability[:-1]
    x_y_list = [list(x_delta_t), list(y_probability)]

    return x_y_list, delta_t_list


def calculate_inter_contact_duration(Data, bins=None):
    r"""Calculates the probabilities for the inter-contact duration

        Calculates the probability for the inter-contact duration delta T of three distinct persons.

        Parameters
        ----------
        Data.interaction: Data
            Contains a dataframe with the tij-data from the data set

        Returns
        -------
        x_y_list: List
            A list with two lists. The first list contains the probabilities for the y-axis and the second list contains
            the inter-contact duration tAC-tAB
        delta_t_list: list
            Contains all the calculated $\delta t$ values for the contact duration of the used data set.

        References
        ----------

        .. [1] Cattuto C, Van den Broeck W, Barrat A, Colizza V, Pinton JF, et al. (2010) Dynamics of Person-to-Person
               Interactions from Distributed RFID Sensor Networks. PLOS ONE


        Examples
        ---------
        For this small example data set we fixed the bins for this function. This is not necessary for the normal usage.

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>               meta_attr_list=attr_list)
        >>> bins = [20, 40, 60, 80]
        >>> x_y_list, delta_t_list = calculate_inter_contact_duration(test_df, bins=bins)
        >>> print(x_y_list[0])
        [0.0, 0.049999999999999996, 0.0]
        >>> print(x_y_list[1])
        [20, 40, 60]

        See Also
        ---------
        face2face.statistics.distribution.calculate_contact_duration
        face2face.statistics.distribution.calculate_triangle_duration

        """

    individuals = list(set(list(Data.interaction.i) + list(Data.interaction.j)))
    delta_t_list = np.array([])
    for ind in individuals:
        time_stamp = Data.interaction[(Data.interaction.i == ind) | (Data.interaction.j == ind)].Time.values
        diff = time_stamp[1:] - time_stamp[:-1]
        delta_t_list = np.append(delta_t_list, diff[diff > 20])

    if bins is None:
        bins = 10**np.linspace(np.log10(min(delta_t_list)), np.log10(max(delta_t_list)), 50)
        x_delta_t, y_probability = np.histogram(delta_t_list, bins=bins, density=True)
    else:
        x_delta_t, y_probability = np.histogram(delta_t_list, bins=bins, density=True)

    y_probability = y_probability[:-1]
    x_y_list = [list(x_delta_t), list(y_probability)]
    return x_y_list, delta_t_list



from face2face.statistics.network_quantities import calculating_number_of_edges_nodes, \
    mean_degree_network, variance_std_network, average_path_length_network, clustering_coefficient


def print_network_measures(network):
    r"""Prints information about the network

        Prints multiple measures as a text output about the given network.

        Parameters
        ----------
        network: networkx Graph
                A graph with a specified degree sequence. Nodes are labeled based on the imported dataset.

        Returns
        -------
        None


        """
    n_nodes, n_edges = calculating_number_of_edges_nodes(network)
    k = mean_degree_network(network)
    var, std = variance_std_network(network)
    d = average_path_length_network(network)
    c = clustering_coefficient(network)

    print("Number of nodes" + " = " + str(n_nodes) + " , " + "Number of edges" + " = " + str(n_edges))
    print("Average Network degree <k>" + " = " + str(k))
    print("Standard deviation of the Network degree" + " = " + str(std))
    print("Average Path Length <d>" + " = " + str(d))
    print("Clustering Coefficient C" + " = " + str(c))

    return None



def check_compatibility_of_versions():

    """Check installed packages needed for this library

       Checks if the installed package versions that are needed for this library contain the needed functions.

       Parameters
       ----------
       None

       Returns
       -------
       dictionary: dict
           Contains the packages needed for this library with the probably missing functions of the packages

       See Also
       ---------
       face2face.compatibility_check.check_compatibility.check_which_methods_are_affected
       """

    dictionary = {}

    import pandas
    pandas_requirements = ["merge", "groupby", "size", "reset_index", "iterrows", "columns", "values", "tail",
                           "read_csv", "rename", "index", "fillna", "replace", "loc", "sort_values", "cumsum",
                           "drop_duplicates", "diff","tolist"]
    pandas_functions = dir(pandas.DataFrame) + dir(pandas) + dir(pandas.Series)
    dictionary["pandas"] = []
    for i in pandas_requirements:
        if i not in pandas_functions:
            dictionary["pandas"].append(i)

    import networkx
    networkx_requirements = ["add_nodes_from", "add_edge", "set_node_attributes", "degree", "nodes", "edges",
                             "attribute_mixing_matrix",
                             "assortativity", "configuration_model", "copy", "remove_edges_from"]

    networkx_functions = dir(networkx) + dir(networkx.Graph)
    dictionary["networkx"] = []
    for i in networkx_requirements:
        if i not in networkx_functions:
            dictionary["networkx"].append(i)

    import numpy
    numpy_requirements = ["tolist", "linspace", "array", "append", "histogram", "shuffle", "ravel", "log10"]

    numpy_functions = dir(numpy) + dir(numpy.ndarray) + dir(numpy.random)
    dictionary["numpy"] = []
    for i in numpy_requirements:
        if i not in numpy_functions:
            dictionary["numpy"].append(i)

    import powerlaw

    powerlaw_requirements = ["Fit", "generate_random", "distribution_compare"]

    powerlaw_functions = dir(powerlaw) + dir(powerlaw.Distribution) + dir(powerlaw.Fit)
    dictionary["powerlaw"] = []
    for i in powerlaw_requirements:
        if i not in powerlaw_functions:
            dictionary["powerlaw"].append(i)

    import matplotlib.pyplot as plt
    from matplotlib.text import Annotation
    matplotlib_requirements = ["subplots", "set_title", "set_text", "title", "twiny", "bar", "set_ylabel", "set_xlabel",
                               "set_xlim","tick_params", "axhline", "axhspan", "axvline", "text", "legend", "boxplot",
                               "xticks","ylabel", "xlabel", "show","tight_layout","plot"]

    matplotlib_functions = dir(plt) + dir(plt.Axes) + dir(Annotation)
    dictionary["matplotlib"] = []
    for i in matplotlib_requirements:
        if i not in matplotlib_functions:
            dictionary["matplotlib"].append(i)

    from scipy.stats import norm
    scipy_function = dir(norm)
    if "sf" not in scipy_function:
        dictionary["scipy"] = "sf"

    import random
    random_function = dir(random.Random)
    if "shuffle" not in random_function:
        dictionary["random"] = "shuffle"

    import seaborn
    seaborn_function = dir(seaborn)
    if "heatmap" not in seaborn_function:
        dictionary["seaborn"] = "heatmap"

    import math
    math_requriements = ["log","sqrt"]
    math_function = dir(math)
    dictionary["math"] = []
    for i in math_requriements:
        if i not in math_function:
            dictionary["math"].append(i)

    import collections
    collections_function = dir(collections)
    if "Counter" not in collections_function:
        dictionary["collections"] = "Counter"

    return dictionary


#dictionary2 = check_compatibility_of_versions()
#print(dictionary2)


def check_which_methods_are_affected(dict):

    """Check for affected library functions

       Checks which library functions are directly affected by missing functions of the installed version of the
       packages. Based on that the user can decide if he needs this functions and therefore another version of a
       package.

       Parameters
       ----------
       dict: dict
           Contains the packages needed for this library with the probably missing functions of the packages

       Returns
       -------
       None

       See Also
       ---------
       face2face.compatibility_check.check_compatibility.check_compatibility_of_versions
       """

    import pandas
    liste = []
    df = pandas.read_csv("Functions_for_functions.csv", sep=';', engine='python')
    for column in df:
        for key, value in dict.items():
            for i in value:
                if i in df[column].tolist():
                    liste.append(str(column))

    liste = list(set(liste))

    dictionary3 = {
        "create_network" : ["create_network","hopping","sliding","event","replace_float_2"],
        "load_all_data" : ["Data"],
        "average_degree" : ["new_avg_degree","avg_degree","new_group_degree","group_degree"],
        "calculating_multiple_measurements_for_a_network" : ["number_edges_nodes","mean_degree","average_path"],
        "check_distribution" : ["search_best_fit","get_best_fit","tupels_comb"],
        "distribution" : ["contact_duration","triangle_duration","inter_contact"],
        "null_modell" : ["bonferroni","mapping","other_model","remove_self_loops","to_simple_graph"],
        "plot_average_degree" : ["plot_degree"],
        "plot_contact_matrix_heatmap" : ["plot_heatmap"],
        "plot_histogram_null_model" : ["plot_diff_scales", "plot_null_model"],
        "plot_probability_distribution" : ["plot_contact_duration","plot_triangle_duration","plot_inter_duration"]}

    next_list = []
    for key, value in dictionary3.items():
        for i in liste:
            if i in value:
                next_list.append(key)

    next_list = list(set(next_list))

    print("Affected functions are:")
    for i in liste:
        print(i)

    print("Affected methods are: ")
    for i in next_list:
        print(i)

#check_which_methods_are_affected(dictionary2)


import matplotlib.pyplot as plt


def plot_avg_degree_group(attr_degree_list, measures):
    r"""Plot boxplot for the average degrees

        Plots the average degree for every attribute in a boxplot. It also shows the average over all degrees and the
        standard deviation.

        Parameters
        ----------
        attr_degree_list : list
            Contains lists with all degrees for all attribute values. You can get this list with the "group_list_degree"
            function.
        measures : list
            Contains the global average degree, the variance and the standard deviation over all attributes.
            You can get this values by using the "global_avg_var_std"- function.

        Returns
        -------
        None

        References
        ------------
        .. [1] Génois, Mathieu & Zens, Maria & Lechner, Clemens & Rammstedt, Beatrice & Strohmaier, Markus. (2019).
               Building connections: How scientists meet each other during a conference.

        See Also
        ----------
        face2face.statistics.average_degree.group_list_degree
        face2face.statistics.average_degree.global_avg_var_std

    """

    x_list = []
    label_list = []
    for i in attr_degree_list:
        x_list.append(i[2])
        label_list.append(str(i[0]) + " " + str(str(i[1])))

    fig, ax = plt.subplots()
    ax.set_title('Average degree for the attributes of the different groups')
    plt.axhline(measures[0], color='black', linestyle='solid', label='AvgMean', linewidth=1)
    plt.axhspan(measures[0], measures[0] - measures[1] * (1 / 3), color="lightcoral")
    plt.axhspan(measures[0] - measures[1] * (1 / 3), measures[0] - measures[1] * (2 / 3), color="indianred")
    plt.axhspan(measures[0] - measures[1] * (2 / 3), measures[0] - measures[1], color="brown")

    plt.axhspan(measures[0], measures[0] + measures[1] * (1 / 3), color="lightcoral")
    plt.axhspan(measures[0] + measures[1] * (1 / 3), measures[0] + measures[1] * (2 / 3), color="indianred")
    plt.axhspan(measures[0] + measures[1] * (2 / 3), measures[0] + measures[1], color="brown")
    ax.boxplot(x_list, vert=True, labels=label_list, showfliers=False, notch=False)
    plt.xticks(rotation=90)
    plt.ylabel("Average Degree")
    plt.xlabel("Attributes from the different groups")
    plt.show()

    return None



from setuptools import setup, Extension
with open("README.rst", "r") as fh:
    long_description = fh.read()
DEPENDENCIES = ['numpy', 'pandas', 'matplotlib', 'networkx', 'scipy',
                'seaborn', 'powerlaw']
packages = ["face2face",
            "face2face.imports",
            "face2face.statistics",
            "face2face.visualization",
	    "face2face.compatibility_check"
            ]
setup(name="face2face",
      version="0.1.1.1.5 Alpha",
      packages=packages,
      description="Library with basic social science functions",
      url="https://gesiscss.github.io/face2face/",
      #python_requires=">=3.6.1",
      long_description=long_description,
      install_requires=DEPENDENCIES,
      package_data = {"face2face": ["data/*/*.dat"]}
      )


import matplotlib.pyplot as plt

plt.rcParams.update(plt.rcParamsDefault)


def plot_contact_duration(list_x_y, color_list, marker_list, label_list):
    r"""Plots the probability distribution of the contact duration

        Plots the values of the probability on the y-axis and the values of the contact duration Delta T on the x-axis.

        Parameters
        ----------
        list_x_y : list
            Contains the output of the "calculate_contact_duration"-function, which is a list of the y-values
            (probabilities) and the x-values (contact durations). You can add multiple lists of the
            "calculate_contact_duration"-function in the list_x_y to compare them in the graph.

        color_list: list
            Should contain a color for every list in list_x_y to define the color of the markers for the visualisation.

        marker_list: list
            Should contain a marker for every list in list_x_y to define the shape of the data points.

        label_list: list
            Should contain a string label for every list in list_x_y to create a comprehensible legend.

        Returns
        -------
        None

        References
        ----------
        .. [1] Cattuto C, Van den Broeck W, Barrat A, Colizza V, Pinton JF, et al. (2010) Dynamics of Person-to-Person
               Interactions from Distributed RFID Sensor Networks. PLOS ONE 5(7): e11596.
               https://doi.org/10.1371/journal.pone.0011596

        See Also
        ---------
        face2face.statistics.distribution.calculate_contact_duration
        """
    for i in range(len(list_x_y)):
        plt.plot(list_x_y[i][1], list_x_y[i][0], marker_list[i], color=color_list[i], label=label_list[i])
        plt.ylabel("Probability ($\Delta$t)")
        plt.xlabel("Contact duration $\Delta$t (seconds)")
        plt.yscale("log")
        plt.xscale("log")
        plt.title("Probability Distribution of duration of contacts")
        x_vertical_lines = [[60, 70,"1 minute"], [1800,2000,"30 minutes"]]
        for xc in x_vertical_lines:
            plt.axvline(x=xc[0],  color='k', linestyle='--')
            plt.text(x=xc[1], y=min(list_x_y[0][0]) + 0.0005, s=xc[2], rotation=90)
        plt.legend(loc="upper right")
    plt.show()

    return None


def plot_triangle_duration(list_x_y, color_list, marker_list, label_list):
    r"""Plots the probability distribution of the triangle duration

        Plots the values of the probability on the y-axis and the values of the triangle duration Delta T on the x-axis.

        Parameters
        ----------
        list_x_y : list
            Contains the output of the "calculate_triangle_duration"-function, which is a list of the y-values
            (probabilities) and the x-values (triangle durations). You can add multiple lists of the
            "calculate_triangle_duration"-function in the list_x_y to compare them in the graph.

        color_list: list
            Should contain a color for every list in list_x_y to define the color of the markers for the visualization.

        marker_list: list
            Should contain a marker for every list in list_x_y to define the shape of the data points.
        label_list: list
            Should contain a string label for every list in list_x_y to create a comprehensible legend.

        Returns
        -------
        None

        References
        ----------
        .. [1] Cattuto C, Van den Broeck W, Barrat A, Colizza V, Pinton JF, et al. (2010) Dynamics of Person-to-Person
               Interactions from Distributed RFID Sensor Networks. PLOS ONE 5(7): e11596.
               https://doi.org/10.1371/journal.pone.0011596

        See Also
        ---------
        face2face.statistics.distribution.calculate_triangle_duration

        """

    for i in range(len(list_x_y)):
        plt.plot(list_x_y[i][1], list_x_y[i][0], marker_list[i], color=color_list[i], label=label_list[i])
        plt.ylabel("Probability($\Delta$t)")
        plt.xlabel("Triangle duration $\Delta$t (seconds)")
        plt.yscale("log")
        plt.xscale("log")
        plt.title("Probability Distribution of the duration of triangles")
        x_vertical_lines = [[60, 70, "1 minute"], [1800, 2000, "30 minutes"]]
        for xc in x_vertical_lines:
            plt.axvline(x=xc[0], color='k', linestyle='--')
            plt.text(x=xc[1], y=min(list_x_y[0][0]) + 0.0005, s=xc[2], rotation=90)
        plt.legend(loc="upper right")
    plt.show()

    return None


def plot_inter_contact_duration(list_x_y, color_list, marker_list, label_list):
    r"""Plots the probability distribution for the inter-contact duration

        Plots the values of the probability on the y-axis and the values of the inter-contact duration Delta T
        on the x-axis.

        Parameters
        ----------
        list_x_y : list
            Contains the output of the "calculate_inter_contact_duration"-function, which is a list of the y-values
            (probabilities) and the x-values (inter-contact durations). You can add multiple lists of the
            "calculate_inter_contact_duration"-function in the list_x_y to compare them in the graph.

        color_list: list
            Should contain a color for every list in list_x_y to define the color of the markers for the visualisation.

        marker_list: list
            Should contain a marker for every list in list_x_y to define the shape of the data points.
        label_list: list
            Should contain a string label for every list in list_x_y to create a comprehensible legend.

        Returns
        -------
        None

        References
        ----------
        .. [1] Cattuto C, Van den Broeck W, Barrat A, Colizza V, Pinton JF, et al. (2010) Dynamics of Person-to-Person
               Interactions from Distributed RFID Sensor Networks. PLOS ONE 5(7): e11596.
               https://doi.org/10.1371/journal.pone.0011596

        See Also
        ---------
        face2face.statistics.distribution.calculate_inter_contact_duration

        """

    for i in range(len(list_x_y)):
        plt.plot(list_x_y[i][1], list_x_y[i][0],marker_list[i], color=color_list[i], label=label_list[i])
        plt.ylabel("Probability ($\Delta$t)")
        plt.xlabel("Inter-contact duration tAC-tAB (seconds)")
        plt.yscale("log")
        plt.xscale("log")
        plt.title("Probability Distribution of the time intervals between consecutive contacts")
        x_vertical_lines = [[60, 70, "1 minute"], [1800, 2000, "30 minutes"]]
        for xc in x_vertical_lines:
            plt.axvline(x=xc[0], color='k', linestyle='--')
            plt.text(x=xc[1], y=min(list_x_y[0][0]) + 0.0005, s=xc[2], rotation=90)
        plt.legend(loc="upper right")
    plt.show()

    return None



import math

from face2face.imports.load_all_data import Data


def calculating_number_of_edges_nodes(network):
    r"""Calculate number of edges/nodes

        Calculating the amount of edges and nodes for a given network

        Parameters
        ----------
        network: networkx Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported data set.

        Returns
        -------
        number_of_nodes: int
            Contains the number of nodes.
        number_of_edges: int
            Contains the number of edges.

        Examples
        ---------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>               meta_attr_list=attr_list)
        >>> test_network = create_network_from_data(test_df)
        >>> number_of_nodes, number_of_edges = calculating_number_of_edges_nodes(test_network)
        >>> print(number_of_nodes)
        10
        >>> print(number_of_edges)
        10

        """
    number_of_nodes = len(network.nodes)
    number_of_edges = len(network.edges)

    return number_of_nodes, number_of_edges


def mean_degree_network(network):
    r"""Calculate average degree

        Calculating the average degree of a given random network

        .. math::

            \langle k \rangle = \frac{2 * \langle L \rangle}{N}

        Parameters
        ----------
        network: networkx Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported dataset.

        Returns
        -------
        k: float
            Contains the average degree for a network.

        References
        ----------
        .. [1] Barabasi, Albert-Laszlo. (2013). Network science. Philosophical transactions. Series A, Mathematical,
               physical, and engineering sciences. 371. 20120375. 10.1098/rsta.2012.0375.

        Examples
        --------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>               path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>               meta_attr_list=attr_list)
        >>> test_network = create_network_from_data(test_df)
        >>> k = mean_degree_network(test_network)
        >>> print(k)
        2

        """
    k = (2 * len(network.edges))/len(network.nodes)

    return k


def variance_std_network(network):
    r"""Calculate variance and standard deviation

        Calculating the variance and the standard deviation of the degrees of a given network

        .. math::

            \sigma^2 = \frac{1}{n} \sum_{i=1}^{n} (x_i - \bar{x})

        .. math::

            \sigma = \sqrt{\sigma^2} = \sqrt{\frac{1}{n-1} \sum_{i=1}^{n} (x_i - \bar{x})^2}

        Parameters
        ----------
        network: networkx Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported data set.

        Returns
        -------
        var: float
            Contains the variance of the network degree
        std: float
            Contains the standard deviation of the network degree

        Examples
        ---------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>                meta_attr_list=attr_list)
        >>> test_network = create_network_from_data(test_df)
        >>> variance, standard_deviation = variance_std_network(test_network)
        >>> print(variance)
        0.2
        >>> print(standard_deviation)
        0.4472135954999579

        """
    var = 0
    k = mean_degree_network(network)
    for i in network.degree:
        var += (i[1] - k)**2
    var = var/len(network.degree)
    std = var**(1/2)

    return var, std


def average_path_length_network(network):
    r"""Calculate the average path length

        Calculating the average path length <d> of a given network

        .. math::

            d_{max} \approx \frac{\ln N}{\ln \langle k \rangle}

        Parameters
        ----------
        network: networkx Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported data set.

        Returns
        -------
        d: float
            Contains the Average Path Length

        References
        ----------
        .. [1] Barabasi, Albert-Laszlo. (2013). Network science. Philosophical transactions. Series A, Mathematical,
               physical, and engineering sciences. 371. 20120375. 10.1098/rsta.2012.0375.

        Examples
        --------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>                meta_attr_list=attr_list)
        >>> test_network = create_network_from_data(test_df)
        >>> d = average_path_length_network(test_network)
        >>> print(d)
        3.3219280948873626

        """

    n_nodes, n_edges = calculating_number_of_edges_nodes(network)
    k = mean_degree_network(network)
    d = math.log(n_nodes)/math.log(k)

    return d


def clustering_coefficient(network):
    r"""Calculate the global Clustering Coefficient

        Calculating the global Clustering Coefficient C for a given network

        .. math::

            \langle L_i \rangle = p \frac{k_i(k_i - 1)}{2}

        .. math::

            C_i = \frac{2 \langle L_i \rangle}{k_i(k_i - 1)} = p = \frac{\langle k \rangle}{N}

        Parameters
        ----------
        network: networkx Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported data set.

        Returns
        -------
        c: float
            Contains the Clustering Coefficient for the network.

        References
        ----------
         .. [1] Barabasi, Albert-Laszlo. (2013). Network science. Philosophical transactions. Series A, Mathematical,
                physical, and engineering sciences. 371. 20120375. 10.1098/rsta.2012.0375.


        Examples
        --------

        >>> attr_list = ["ID", "Age", "Sex"]
        >>> test_df = Data(path_tij="face2face/data/Test/tij_test.dat", separator_tij="\t",
        >>>                path_meta="face2face/data/Test/meta_test.dat", separator_meta="\t",
        >>>                meta_attr_list=attr_list)
        >>> test_network = create_network_from_data(test_df)
        >>> C = clustering_coefficient(test_network)
        >>> print(C)
        0.2

        """
    n_nodes, n_edges = calculating_number_of_edges_nodes(network)
    k = mean_degree_network(network)
    c = k / n_nodes

    return c



import random

import networkx as nx
import numpy as np
import scipy


from face2face.imports.load_all_data import Data
from face2face.imports.create_network import create_network_from_data

np.set_printoptions(precision=16)


def check_bonferroni_correction(contact_matrix):

    r"""Apply the bonferroni correction on the contact matrix

        Calculates the p_values of the contact matrix and applies the bonferroni correction on the contact matrix

        Parameters
        ----------
        contact_matrix: list
            A matrixlike List of lists that contains the z-scores for a given attribute.

        Returns
        -------
        annotation_list: list
            A matrixlike list of lists that contains the annotated z-scores for a given attribute. They are annotated
            with a * if they are too small in terms of the bonferroni correction.

        Notes
        -----
        The annotation_list contains the z-scores of the contact matrix and adds a * at every z-score that is an outlier

        References
        ----------
        .. [1] Génois, Mathieu & Zens, Maria & Lechner, Clemens & Rammstedt, Beatrice & Strohmaier, Markus. (2019).
               Building connections: How scientists meet each other during a conference.

        Examples
        ---------
        >>> cm = np.array([[1, 1, 3.4], [1, 1, 1], [3.4, 1, 1]])
        >>> annotated_cm = check_bonferroni_correction(cm)
        >>> print(annotated_cm)
        [['1.0', '1.0', '3.4*'], ['1.0', '1.0', '1.0'], ['3.4*', '1.0', '1.0']]

        See Also
        ---------
        face2face.statistics.null_modell.configuration_model_label_z_score_mixing_matrix
        face2face.statistics.null_modell.shuffle_label_z_score_mixing_matrix

        """
    alpha = 0.01
    p_values = scipy.stats.norm.sf(abs(contact_matrix))
    k = len(contact_matrix) - 1
    p_value = 1-(1-alpha)**(1/k)
    annotation_list = []
    for i in range(len(p_values)):
        annot_list = []
        for j in range(len(p_values[i])):
            if p_values[i][j] < p_value:
                annot_z_score = str(round(contact_matrix[i][j], 2)) + '*'
                annot_list.append(annot_z_score)
            else:
                annot_z_score = round(contact_matrix[i][j], 2)
                annot_list.append(annot_z_score)
        annotation_list.append(annot_list)
    annotation_list = np.array(annotation_list)

    return annotation_list


def mapping_function(Data, label="type"):

    r"""Creates a mapping

        Creates a mapping dictionary to create the attribute mixing matrix dynamic,
        independent of the amount of attributes.

        Parameters
        ----------
        Data: Data
            Data Object that contains Tij- and Metadata for a data set.
        label: str
            A string that tells the function which attribute should be mapped.

        Returns
        -------
        mapping: dict
            A dictionary that contains the dimensions of the given attribute.

        Examples
        ---------
        Based on the amount of attribute values the used attribute has, you have to create a mapping for the
        null model functions.

        >>> mapping = mapping_function(test_df, label="Age")
        >>> print(mapping)
        {0: 0, 1: 1, 2: 2}

        >>> mapping = mapping_function(test_df, label="Sex")
        >>> print(mapping)
        {0: 0, 1: 1}

        See Also
        ---------
        face2face.statistics.null_modell.configuration_model_label_z_score_mixing_matrix
        face2face.statistics.null_modell.shuffle_label_z_score_mixing_matrix

        """
    df_meta_nan = Data.metadata.fillna("NaN")
    attr_list = set(df_meta_nan[label])
    if "NaN" in attr_list:
        attr_list.remove("NaN")
    mapping = {}
    for i in range(len(attr_list)):
        mapping[i] = i

    return mapping


def configuration_model_label_z_score_mixing_matrix(Data, runs=1000, label="type", shuffle_label=False,
                                                    force_simple_graph=False, seed_config_mat=None, seed_label=None):

    r"""Creates a contact matrix based on the configuration model

        Creates a contact matrix with z-scores based on the chosen attribute. You can assume randomized attributes and/
        or randomized degrees in the null model.

        Parameters
        ----------
        Data: Data
            Data Object that contains Tij- and Metadata for a data set.
        runs: int
            The amount of times the function should be executed. It's a heuristic approach, so the more the runs the
            better might be the result
        label: str
            A string that tells the function for which attribute the contact matrix should be made.
        shuffle_label: bool
            Gives the option to extend the null model by randomizing the node attributes.
        force_simple_graph: bool
            Deletes parallel- and selfedges that can occur by using the networkx "configuration_model"-function if True.
        seed_config_mat : List, default None
            Allows to create reproducible configuration models for a reproducible output. This parameters is basically
            just for applying tests.
        seed_label : list, default None
            Allows to create reproducible "randomized" labels for a reproducible output. This parameters is basically
            just for applying tests.

        Returns
        -------
        contact_matrix : list
             A matrixlike List of lists that contains the z-scores for a given attribute

        References
        ----------
        .. [1] Génois, Mathieu & Zens, Maria & Lechner, Clemens & Rammstedt, Beatrice & Strohmaier, Markus. (2019).
               Building connections: How scientists meet each other during a conference.

        Examples
        ---------
        >>> contact_matrix = configuration_model_label_z_score_mixing_matrix(test_network, test_df, runs=1000,
        >>>                                                                  label="Age", shuffle_label=True,
        >>>                                                                  force_simple_graph=True)
        >>> print(contact_matrix)
        [[1.0320936930842797, -0.7717976357301974, -0.5],
        [-0.7717976357301974, -0.0667601413575786, -1.2927763604862383],
        [-0.5, -1.2927763604862383, 2.632147318581194]]

        See Also
        ---------
        face2face.statistics.null_modell.shuffle_label_z_score_mixing_matrix
        """

    network = create_network_from_data(Data, replace_attr=True, label=label)
    mapping = mapping_function(Data, label)
    data_mixing_matrix = nx.assortativity.attribute_mixing_matrix(network, label, mapping=mapping)
    degree_sequence = [v[1] for v in network.degree]
    type_sequence = [network.nodes[n][label] for n in network.nodes]
    matrices = []
    matrices_abs = []
    for _ in range(runs):
        if seed_config_mat is None:
            null_model = nx.configuration_model(degree_sequence)
        else:
            null_model = nx.configuration_model(degree_sequence, seed=seed_config_mat[_])
        if force_simple_graph:
            null_model = to_simple_graph(null_model)
        if shuffle_label:
            if seed_label is None:
                np.random.shuffle(type_sequence)
            else:
                random.Random(seed_label[_]).shuffle(type_sequence)
        for n, t in zip(null_model.nodes, type_sequence):
            null_model.nodes[n][label] = t
        matrices.append(nx.assortativity.attribute_mixing_matrix(null_model, label, mapping=mapping))
        matrices_abs.append(nx.assortativity.attribute_mixing_matrix(null_model, label, mapping=mapping, normalized=False))

    return (data_mixing_matrix - np.array(matrices).mean(axis=0))/np.array(matrices).std(axis=0), matrices, \
        data_mixing_matrix, matrices_abs


def shuffle_label_z_score_mixing_matrix(Data, runs=1000, label="type", seed_label=None):

    r"""Creates a contact matrix based on a null model with randomized identities.

        Creates a contact matrix with z-scores based on the chosen attribute. You can assume randomized identities.

        Parameters
        ----------
        Data: Data
            Data Object that contains Tij- and Metadata for a data set.
        runs: int
            The amount of times the function should be executed. It's a heuristic approach, so the more the runs the
            better might be the result
        label: str
            A string that tells the function for which attribute the contact matrix should be made.
        seed_label : list, default None
            Allows to create reproducible "randomized" labels for a reproducible output. This parameters is basically
            just for applying tests.

        Returns
        -------
        contact_matrix : list
             A matrixlike List of lists that contains the z-scores for a given attribute and randomized identities.

        References
        ----------
        .. [2] Génois, Mathieu & Zens, Maria & Lechner, Clemens & Rammstedt, Beatrice & Strohmaier, Markus. (2019).
               Building connections: How scientists meet each other during a conference.

        Examples
        ---------
        >>> contact_matrix = shuffle_label_z_score_mixing_matrix(test_network, test_df, runs=1000, label="Age"):
        >>> print(contact_matrix)
        [[1.2247448713915892, -0.44846105565116173, -1.5452456409610384],
        [-0.44846105565116173, 2.0, -3.143958736099446],
        [-1.5452456409610384, -3.143958736099446, 5.719237485832778]]

        See Also
        ---------
        face2face.statistics.null_modell.configuration_model_label_z_score_mixing_matrix

        """

    df = Data

    network = create_network_from_data(df, replace_attr=True, label=label)
    network = network.copy()
    mapping = mapping_function(df, label)
    data_mixing_matrix = nx.assortativity.attribute_mixing_matrix(network, label, mapping=mapping)

    def shuffle_labels(graph):
        if seed_label is None:
            data_group_nodes = [graph.nodes[n][label] for n in graph.nodes]
            np.random.shuffle(data_group_nodes)
        else:
            data_group_nodes = [graph.nodes[n][label] for n in graph.nodes]
            random.Random(seed_label[_]).shuffle(data_group_nodes)
        for n, new_label in zip(graph.nodes, data_group_nodes):
            graph.nodes[n][label] = new_label
        return graph
    matrices = []
    matrices2 = []
    for _ in range(runs):
        shuffle_labels(network)
        matrices.append(nx.assortativity.attribute_mixing_matrix(network, label, mapping=mapping))
        matrices2.append(nx.assortativity.attribute_mixing_matrix(network, label, mapping=mapping, normalized=False))

    return (data_mixing_matrix - np.array(matrices).mean(axis=0))/np.array(matrices).std(axis=0), matrices2


def remove_self_loops(graph):

    r"""Removes self-loops that occur by using the "configuration_model"-function

        Removes self-loops that occur by using the "configuration_model"-function from networkx.

        Parameters
        ----------
        graph : networkX Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported data set. Graph might
            contain self loops

        Returns
        ---------
        graph : networkX Graph
            A graph with a specified degree sequence. Nodes are labeled based on the imported data set. Graph contains
            no selfloops anymore.

        Examples
        ---------
        The :func:`networkx.configuration_model` function which is being used for the configuration model can lead to
        self- and parallel edges. As you can see here the function :func:`remove_self_loops` filters self loops out of
        the network.

        >>> degree_sequence_1 = [v[1] for v in test_network.degree]
        >>> test_model = nx.configuration_model(degree_sequence_1)
        >>> print(test_model.edges)
        [(0, 1, 0), (1, 1, 0), (2, 6, 0), (2, 3, 0), (3, 4, 0), (4, 8, 0), (5, 5, 0), (6, 9, 0), (7, 8, 0), (7, 9, 0)]
        >>> remove_self_loops(test_model)
        >>> print(test_model.edges)
        [(0, 1, 0), (2, 6, 0), (2, 3, 0), (3, 4, 0), (4, 8, 0), (6, 9, 0), (7, 8, 0), (7, 9, 0)]

         See Also
         ---------
         face2face.statistics.null_modell.configuration_model_label_z_score_mixing_matrix

         """
    self_loops = []
    for e in graph.edges:
        if e[0] == e[1]:
            self_loops.append(e)
    graph.remove_edges_from(self_loops)
    return graph


def to_simple_graph(graph):
    graph = nx.Graph(graph)
    graph = remove_self_loops(graph)
    return graph


